// $Id: log.cxx 3866 2005-09-07 10:44:22Z jz $

#include "zschimmer.h"
#include "log.h"

using namespace std;


namespace zschimmer {

//------------------------------------------------------------------------------------------extern

System_mutex*       Log_ptr::_system_mutex_ptr  = NULL;
ostream**           Log_ptr::_stream_ptr        = NULL;
uint                Log_ptr::_indent_tls_index  = 0;
Log_categories      log_categories;

//--------------------------------------------------------------------------------------------const

static Error_code_text error_codes[] =
{
    { "Z-LOG-001", "Ungltiger Log-Level $1: Nur error, warn, info und debug1 bis debug9 sind mglich" },
    { NULL       , NULL }
};

//-------------------------------------------------------------------------------------------Z_INIT

Z_INIT( log )
{
    add_error_code_texts( error_codes );
}

//-----------------------------------------------------------------------------------make_log_level

Log_level make_log_level( const string& name )
{
    Log_level log_level = log_debug9;

    if( name == "error"   )  log_level = log_error;
    else
    if( name == "warn"
     || name == "warning" )  log_level = log_warn;
    else                     
    if( name == "info"    )  log_level = log_info;
    else
    if( name == "debug"   )  log_level = log_debug;
    else
    if( name == "unknown" )  log_level = log_unknown;
    else
    if( strncmp(name.c_str(),"debug",5) == 0 )
    {
        try {
            log_level = (Log_level)-(int)as_uint( name.c_str() + 5 );
        }
        catch( const Xc& ) { throw_xc( "Z-LOG-001", name ); }
    }
    else
    {
        try {
            log_level = (Log_level)as_int( name );
            if( log_level != log_unknown  
             && ( log_level < log_debug9  ||  log_level > log_error ) )  throw_xc( "Z-LOG-001", name );  //log_level = log_error;
        }
        catch( const Xc& ) { throw_xc( "Z-LOG-001", name ); }
    }

    return log_level;
}

//--------------------------------------------------------------------------------name_of_log_level

string name_of_log_level( Log_level level )
{
    if( level >= log_debug9  &&  level <= log_debug1 )  return "debug" + as_string( -level );
    
    switch( level )
    {
        case log_info:    return "info";
        case log_warn:    return "warning";
        case log_error:   return "error";
        case log_fatal:   return "fatal";
        case log_unknown: return "unknown";
        default:          return as_string( level );
    }
}

//------------------------------------------------------------------------------Delegated_log::log2

void Delegated_log::log2( Log_level level, const string& prefix, const string& line, Has_log* prefix_log )   
{ 
    if( this != NULL  &&  level < _log_level )  return;
    
    if( this != NULL  &&  _log )  
    {
        _log->log2( level, prefix, line, prefix_log? prefix_log : this ); 
    }
    else
    {
        // Mglicherweise ist this == null.

        Z_LOG( line << '\n' );
        fprintf( stderr, "[%s] %s\n", prefix.c_str(), line.c_str() );
    }
}

//---------------------------------------------------------------------------og_ptr::Indent::indent

void Log_ptr::Indent::indent()
{
    indent_( +1 );

    Log_ptr().flush();
}

//--------------------------------------------------------------------------og_ptr::Indent::~Indent

Log_ptr::Indent::~Indent()
{
    indent_( -1 );
}

//--------------------------------------------------------------------------og_ptr::Indent::indent_

void Log_ptr::Indent::indent_( int plus_or_minus )
{
    if( _indent_tls_index )
    {
#       if defined Z_WINDOWS

            size_t indent = (size_t)TlsGetValue( _indent_tls_index );
            indent += plus_or_minus;
            TlsSetValue( _indent_tls_index, (void*)indent ); 

#       endif
    }
}

//----------------------------------------------------------------------------------og_ptr::Log_ptr

Log_ptr::Log_ptr()
:
    _stream( _stream_ptr? *_stream_ptr : NULL ) 
{ 
    if( _stream )  enter_mutex( _system_mutex_ptr ); 
}

//----------------------------------------------------------------------------------og_ptr::Log_ptr

Log_ptr::Log_ptr( const char* category )
:
    _stream( log_category_is_set(category) && _stream_ptr? *_stream_ptr : NULL )
{ 
    if( _stream )
    {
        enter_mutex( _system_mutex_ptr ); 

        if( category  &&  category[0] )  *_stream << '{' << category << "} ";
    }
}

//--------------------------------------------------------------------------------Log_ptr::~Log_ptr

Log_ptr::~Log_ptr()
{ 
    if( _stream )
    {
        _stream->flush(); 
        leave_mutex( _system_mutex_ptr ); 
    }
}

//-----------------------------------------------------------------------------------Log_categories
/*
    Beispiel:

    Der Benutzer stellt (z.B. in einer .ini-Datei) ein:
    
        log_what = misc.profile misc.java.exceptions objectserver.* objectserver.recv-
    

    Das fhrt zu den Aufrufen:

        set( "misc.profile"         , true );
        set( "misc.java.exception"  , true );
        set( "objectserver.*"       , true );
        set( "objectserver.recv"    , false );


    Das ergibt die _map:

        "misc.profile"         ._value = true
        "misc.java.exception"  ._value = true 
        "objectserver.*"       ._value = true 
        "objectserver.recv"    ._value = false


    Das Programm setzt Voreinstellungen (vor oder nach den set()-Aufrufen):

        set_default( "misc.profile.get"   , true );
      //set_default( "misc.profile.include, false );        // include ausnehmen

        set_default( "misc.java.call"     , true );
        set_default( "misc.java.exception", true );
        // misc.java => misc.java.call misc.java.exception

      //set_default( "objectserver.send"    , false );
      //set_default( "objectserver.recv"    , false );
        set_default( "objectserver.call"    , true );
        set_default( "objectserver.callback", true );
        // objectserver => objectserver.call objectserver.callback


    Das ergibt die _default_map:

        "misc.profile.get"   , true );
      //"misc.profile.include, false );        // include ausnehmen

        "misc.java.call"     , true );
        "misc.java.exception", true );
        // misc.java => misc.java.call misc.java.exception

      //"objectserver.send"    , false );
      //"objectserver.recv"    , false );
        "objectserver.call"     ._default_value, true );
        "objectserver.callback" ._default_value =  true );


    misc.profile.get
    misc.profile.include
    misc.profile.include misc.profile.get-
    misc.com

    log.programname
    log.virtual_size

    scheduler.wait
    scheduler.sockets

    java.calls
    java.stacktrace
    java.options

    objectserver.call

    -------------------

    log_categories.add( "misc.profile !log.programmname java.calls" );
    
    log_categories.set_default( "misc.profile.get" );


    if( log_categories( "misc.profile.include" ) ) ...;

*/

//--------------------------------------------------------------------og_categories::Log_categories

Log_categories::Log_categories()
: 
    _mutex( "log_categories", Mutex::kind_recursive_dont_log ) 
{  
    _valid = true; 
}

//---------------------------------------------------------------------------Log_categories::is_set

bool Log_categories::is_set( const char* name )
{
    return Log_ptr::logging()  &&  ( name == NULL  ||  name[0] == '\0'  ||  is_set( string( name ) ) );
}

//---------------------------------------------------------------------------Log_categories::is_set

bool Log_categories::is_set( const string& name )
{
    if( !_valid    )  return true; // false?
    if( _all       )  return true;                // Fr set_all(), kann von Signal-Handler (z.B. SIGUSR1) gesetzt werden.
    if( name == "" )  return true;

    bool result = false;

    Z_MUTEX( _mutex )
    {
        Map::iterator it = _map.find( name );
        if( it == _map.end() ) 
        {
            Map::iterator it_all;

            size_t point = name.find( '.' );  
            if( point != string::npos )
            {
                string all_childs = name.substr( 0, point+1 );
                all_childs += '*';     // parent.*


                it_all = _map.find( all_childs );
                if( it_all != _map.end() )  result = it_all->second._value;
            }

            it_all = _map.find( "all" );
            if( it_all != _map.end() )  result = it_all->second._value; 

            _map[ name ]._value = result;       // Eintrag merken, damit's beim nchsten Mal gleich gefunden wird.
        }
        else
        {
            result = !it->second._is_local_default  &  it->second._value;
        }
    }

    return result;
}

//------------------------------------------------------------------------------Log_categories::set

void Log_categories::set_multiple( const string& names )
{
    if( names.empty() )  return;

    if( names.find( ' ' ) != string::npos )
    {
        // Ohne regexp auskommen: vector<string> names_vector = vector_split( "[ \t\r\n]+", names );
        //for( uint i = 0; i < names_vector.size(); i++ )  set( names_vector[i] );

        const char* p = names.c_str();
        string      word;

        while( *p )
        {
            while( isspace( (uint)*p ) )  p++;
            word.erase();
            while( *p  &&  !isspace( (uint)*p ) )  word += *p++;
            set( word );
        }
    }
    else
    {
        set( names );
    }
}

//------------------------------------------------------------------------------Log_categories::set

void Log_categories::set( const string& name )
{
    if( !_valid )  return;


    if( name.empty() )  return;

    if( *name.rbegin() == '-' )  set( name.substr( 0, name.length() - 1 ), false );
                           else  set( name, true );
}

//------------------------------------------------------------------------------Log_categories::set

void Log_categories::set( const string& name, bool value )
{
    if( name.empty() )  return;

    Z_MUTEX( _mutex )
    {
        _modified_counter++;

        //Entry& entry = string_ends_with( name, ".*" )? _map[ name.substr( 0, name.length() - 1 ) ]      // Nur "name.", ohne Stern, damit is_set() schneller ist
        //                                             : _map[ name ];
        Entry& entry = _map[ name ];

        // Merken fr as_string()
        if( !_categories_string.empty() )  _categories_string += ' ';
        _categories_string += name;
        if( !value )  _categories_string += '-';


        entry._is_global_default = false;
        entry._is_local_default = false;
        entry._value = value;


        // Lokale Defaults bernehmen

        string prefix = name + ".";
        Z_FOR_EACH( Map, _map, e )
        {
            if( string_begins_with( e->first, prefix ) )  e->second._is_local_default = false;
        }
    }
}

//----------------------------------------------------------------------Log_categories::set_default

void Log_categories::set_global_default( const string& name, bool value )
{
    Z_MUTEX( _mutex )
    {
        _modified_counter++;

        Map::iterator e = _map.find( name );
        
        if( e == _map.end() )
        {
            _map[ name ]._is_global_default = true;
            _map[ name ]._value             = value;
        }
    }
}

//----------------------------------------------------------------------Log_categories::set_default

void Log_categories::set_local_default( const string& name, bool value )
{
    Z_MUTEX( _mutex )
    {
        _modified_counter++;

        Map::iterator e = _map.find( name );
        
        if( e == _map.end() )
        {
            _map[ name ]._is_local_default = true;
            _map[ name ]._value            = value;
        }
    /*?
        else
        if( !e->second._is_global_default  &  !e->second._is_local_default )
        {
            int point = name.find( '.' );
            if( point == string::npos )  return;

            e = _map.find( name.substr( name, 0, point ) );
            if( e != _map.end()  &&  !e->second._is_global_default  &  !e->second._is_local_default ))
            {
                _map[ name ]._value = e->second._value;
            }
        }
    */
    }
}

//------------------------------------------------------------------------------log_category_is_set
// Deklariert in base.h, damit log.h nicht eingezogen werden muss.

bool                        log_category_is_set         ( const char* name )                        { return log_categories.is_set( name ); }
bool                        log_category_is_set         ( const string& name )                      { return log_categories.is_set( name ); }
void                    set_log_category                ( const string& name, bool value )          { log_categories.set( name, value ); }
void                    set_log_category_default        ( const string& name, bool value )          { log_categories.set_global_default( name, value ); }
void                    set_log_category_local_default  ( const string& name, bool value )          { log_categories.set_local_default( name, value ); }
string                      log_categories_as_string    ()                                          { return log_categories.as_string(); }

//-------------------------------------------------------------------------------------------------

} //namespace zschimmer
